@@ -27,6 +27,11 @@
 #include "xklavier_private.h"
 #include "xklavier_private_xkb.h"
 
+#ifdef HAVE_XINPUT
+#include <X11/extensions/XI.h>
+#include <X11/extensions/XInput.h>
+#endif
+
 #ifdef LIBXKBFILE_PRESENT
 
 const gchar **
@@ -46,22 +51,19 @@ xkl_xkb_get_indicators_names(XklEngine * engine)
 gint
 xkl_xkb_pause_listen(XklEngine * engine)
 {
-	XkbSelectEvents(xkl_engine_get_display(engine), XkbUseCoreKbd,
+	XkbSelectEvents(xkl_engine_get_display(engine),
+			xkl_engine_backend(engine, XklXkb, device_id),
 			XkbAllEventsMask, 0);
-/*  XkbSelectEventDetails( xkl_display,
-                         XkbUseCoreKbd,
-                       XkbStateNotify,
-                     0,
-                   0 );
-
-  !!_XklSelectInput( _xklRootWindow, 0 );
-*/
 	return 0;
 }
 
 gint
 xkl_xkb_resume_listen(XklEngine * engine)
 {
+#ifdef HAVE_XINPUT
+	int xitype;
+	XEventClass xiclass;
+#endif
 	/* What events we want */
 #define XKB_EVT_MASK \
          (XkbStateNotifyMask| \
@@ -72,26 +74,37 @@ xkl_xkb_resume_listen(XklEngine * engine)
           XkbNewKeyboardNotifyMask)
 
 	Display *display = xkl_engine_get_display(engine);
-	XkbSelectEvents(display, XkbUseCoreKbd, XKB_EVT_MASK,
-			XKB_EVT_MASK);
+	XkbSelectEvents(display,
+			xkl_engine_backend(engine, XklXkb, device_id),
+			XKB_EVT_MASK, XKB_EVT_MASK);
 
 #define XKB_STATE_EVT_DTL_MASK \
          (XkbGroupStateMask)
 
 	XkbSelectEventDetails(display,
-			      XkbUseCoreKbd,
-			      XkbStateNotify,
-			      XKB_STATE_EVT_DTL_MASK,
+			      xkl_engine_backend(engine, XklXkb,
+						 device_id),
+			      XkbStateNotify, XKB_STATE_EVT_DTL_MASK,
 			      XKB_STATE_EVT_DTL_MASK);
 
 #define XKB_NAMES_EVT_DTL_MASK \
          (XkbGroupNamesMask|XkbIndicatorNamesMask)
 
 	XkbSelectEventDetails(display,
-			      XkbUseCoreKbd,
-			      XkbNamesNotify,
-			      XKB_NAMES_EVT_DTL_MASK,
+			      xkl_engine_backend(engine, XklXkb,
+						 device_id),
+			      XkbNamesNotify, XKB_NAMES_EVT_DTL_MASK,
 			      XKB_NAMES_EVT_DTL_MASK);
+#ifdef HAVE_XINPUT
+	if (xkl_engine_priv(engine, features) | XKLF_DEVICE_DISCOVERY) {
+		DevicePresence(display, xitype, xiclass);
+		XSelectExtensionEvent(display,
+				      xkl_engine_priv(engine, root_window),
+				      &xiclass, 1);
+		xkl_engine_backend(engine, XklXkb, xi_event_type) = xitype;
+	} else
+		xkl_engine_backend(engine, XklXkb, xi_event_type) = -1;
+#endif
 	return 0;
 }
 
@@ -158,7 +171,9 @@ xkl_xkb_load_actual_desc(XklEngine * engine)
 	Status status;
 
 	Display *display = xkl_engine_get_display(engine);
-	XkbDescPtr desc = XkbGetMap(display, KBD_MASK, XkbUseCoreKbd);
+	XkbDescPtr desc = XkbGetMap(display, KBD_MASK,
+				    xkl_engine_backend(engine, XklXkb,
+						       device_id));
 	xkl_engine_backend(engine, XklXkb, actual_desc) = desc;
 	if (desc != NULL) {
 		rv = Success == (status = XkbGetControls(display,
@@ -217,10 +232,10 @@ xkl_xkb_if_cached_info_equals_actual(XklEngine * engine)
 				rv = i < 0;
 			}
 		}
-    /* 
-     * in case of failure, reuse in _XklXkbLoadAllInfo
-     * in case of success - free it
-     */
+		/* 
+		 * in case of failure, reuse in _XklXkbLoadAllInfo
+		 * in case of success - free it
+		 */
 		if (rv) {
 			XkbFreeKeyboard(actual,
 					XkbAllComponentsMask, True);
@@ -310,7 +325,8 @@ xkl_xkb_lock_group(XklEngine * engine, gint group)
 	Display *display = xkl_engine_get_display(engine);
 	xkl_debug(100, "Posted request for change the group to %d ##\n",
 		  group);
-	XkbLockGroup(display, XkbUseCoreKbd, group);
+	XkbLockGroup(display,
+		     xkl_engine_backend(engine, XklXkb, device_id), group);
 	XSync(display, False);
 }
 
@@ -324,16 +340,21 @@ xkl_xkb_get_server_state(XklEngine * engine, XklState * current_state_out)
 	Display *display = xkl_engine_get_display(engine);
 
 	current_state_out->group = 0;
-	if (Success == XkbGetState(display, XkbUseCoreKbd, &state))
+	if (Success ==
+	    XkbGetState(display,
+			xkl_engine_backend(engine, XklXkb, device_id),
+			&state))
 		current_state_out->group = state.locked_group;
 
 	if (Success ==
-	    XkbGetIndicatorState(display, XkbUseCoreKbd,
+	    XkbGetIndicatorState(display,
+				 xkl_engine_backend(engine, XklXkb,
+						    device_id),
 				 &current_state_out->indicators))
 		current_state_out->indicators &=
 		    xkl_engine_backend(engine, XklXkb,
-				       cached_desc)->indicators->
-		    phys_indicators;
+				       cached_desc)->
+		    indicators->phys_indicators;
 	else
 		current_state_out->indicators = 0;
 }
@@ -375,12 +396,14 @@ xkl_xkb_set_indicator(XklEngine * engine, gint indicator_num, gboolean set)
 
 	case XkbIM_NoAutomatic:
 		{
-			if (cached->names->
-			    indicators[indicator_num] != None)
+			if (cached->names->indicators[indicator_num] !=
+			    None)
 				XkbSetNamedIndicator(display,
-						     XkbUseCoreKbd,
-						     cached->names->
-						     indicators
+						     xkl_engine_backend
+						     (engine, XklXkb,
+						      device_id),
+						     cached->
+						     names->indicators
 						     [indicator_num], set,
 						     False, NULL);
 			else {
@@ -440,13 +463,15 @@ xkl_xkb_set_indicator(XklEngine * engine, gint indicator_num, gboolean set)
 					group = i;
 					break;
 				}
-			if (map->
-			    which_groups & (XkbIM_UseLocked |
-					    XkbIM_UseEffective)) {
+			if (map->which_groups & (XkbIM_UseLocked |
+						 XkbIM_UseEffective)) {
 				/* Important: Groups should be ignored here - because they are handled separately! */
 				/* XklLockGroup( group ); */
 			} else if (map->which_groups & XkbIM_UseLatched)
-				XkbLatchGroup(display, XkbUseCoreKbd,
+				XkbLatchGroup(display,
+					      xkl_engine_backend(engine,
+								 XklXkb,
+								 device_id),
 					      group);
 			else {
 				/* Can do nothing. Just ignore the indicator */
@@ -487,12 +512,17 @@ xkl_xkb_set_indicator(XklEngine * engine, gint indicator_num, gboolean set)
 
 		mods = set ? affect : 0;
 
-		if (map->
-		    which_mods & (XkbIM_UseLocked | XkbIM_UseEffective))
-			XkbLockModifiers(display, XkbUseCoreKbd,
+		if (map->which_mods &
+		    (XkbIM_UseLocked | XkbIM_UseEffective))
+			XkbLockModifiers(display,
+					 xkl_engine_backend(engine, XklXkb,
+							    device_id),
 					 affect, mods);
 		else if (map->which_mods & XkbIM_UseLatched)
-			XkbLatchModifiers(display, XkbUseCoreKbd,
+			XkbLatchModifiers(display,
+					  xkl_engine_backend(engine,
+							     XklXkb,
+							     device_id),
 					  affect, mods);
 		else {
 			return TRUE;
@@ -511,6 +541,7 @@ xkl_xkb_init(XklEngine * engine)
 #ifdef LIBXKBFILE_PRESENT
 	gint opcode;
 	gboolean xkl_xkb_ext_present;
+	int xi_opc, xi_event_type, xi_error_code;
 
 	xkl_engine_priv(engine, backend_id) = "XKB";
 	xkl_engine_priv(engine, features) = XKLF_CAN_TOGGLE_INDICATORS |
@@ -548,6 +579,7 @@ xkl_xkb_init(XklEngine * engine)
 		return -1;
 
 	xkl_engine_priv(engine, backend) = g_new0(XklXkb, 1);
+	xkl_engine_backend(engine, XklXkb, device_id) = XkbUseCoreKbd;
 
 	xkl_xkb_ext_present = XkbQueryExtension(display,
 						&opcode,
@@ -558,12 +590,8 @@ xkl_xkb_init(XklEngine * engine)
 								    XklXkb,
 								    error_code),
 						NULL, NULL);
-	if (!xkl_xkb_ext_present) {
-		XSetErrorHandler((XErrorHandler)
-				 xkl_engine_priv(engine,
-						 default_error_handler));
+	if (!xkl_xkb_ext_present)
 		return -1;
-	}
 
 	xkl_debug(160,
 		  "xkbEvenType: %X, xkbError: %X, display: %p, root: "
@@ -587,6 +615,14 @@ xkl_xkb_init(XklEngine * engine)
 		xkl_engine_priv(engine, features) |=
 		    XKLF_MULTIPLE_LAYOUTS_SUPPORTED;
 
+	if (XQueryExtension
+	    (display, "XInputExtension", &xi_opc,
+	     &xi_event_type, &xi_error_code)) {
+		xkl_debug(150, "XInputExtension found (%d, %d, %d)\n",
+			  xi_opc, xi_event_type, xi_error_code);
+		xkl_engine_priv(engine, features) |= XKLF_DEVICE_DISCOVERY;
+	} else
+		xkl_debug(0, "XInputExtension not found\n");
 	return 0;
 #else
 	xkl_debug(160,
